;;; "PONG" in AVR assembly

; V0.2 - show a bouncing ball

#define __SFR_OFFSET 0 // makes "PORTB" etc fit the IN/OUT Port, instead of the memory equivalent
#include <avr/io.h>    // Get the "basic" symbolic names for registers
.global main           ; Interface to the Arduino/C startup
                       ; (therefore?) stackpointer and vector table initialized

; additional hardware defines (The SPI is given)
CSS = PB2               ; Arduino pin 10 (portB, pin 2) is the CSS line
; [TBW] buttons

DispSegs = 2            ; number of display segments (could be between 2 and 6)
RowLow = 1              ; lowest display row# (but is at the top)
RowHig = 8              ; highest             (              bottom)
ColLft = 0              ; leftmost display column
ColRgt = DispSegs*8-1   ; right

// This uses the fact we are preprocessed and thus can "rename" registers
// register assigments/equivalence for variables
#define Xpos r18        // [0..15]
#define Ypos r19        // [1..8]
#define Xdir r5         // 0 = left, 1 = right (ie 0 is decrementing, 1 is incrementing)
#define Ydir r6         // 0 = up, 1 = down
#define Speed r7        // number of deci-seconds between moves, call it "slowness" instead

; register assignmenst/equivalnce for work variables
#define CurLin r10      // current line we're displaying
#define LinBuf r8       // current line segment
#define LinBuf2 r9      //  2nd half, BUT that hardcodes us to 2 seg only

#define Tmp r16         // for single byte stuff
#define Tmp2 r20        // for double byte stuff

;;;;;;;;;;;;;;;;;;;Code;;;;;;;;;;;;;;;;

; ===== send one row to display
SendRow:
   ; Push four bytes to SPI, bracket by LOW pin 10
   ; uses CurLin and LinBuf/2
   cbi   PORTB,CSS       ; Trigger a "new transaction" on MAX7219 (pin 10 LOW)
   mov   r20,CurLin      ; 
   mov   r21,LinBuf     ; 
   rcall SendTwo
   mov   r21,LinBuf2
   rcall SendTwo
   sbi   PORTB,CSS       ; pin 10 HIGH again
   ret

; ==== send one byte-pair to display (not a transaction!) Values in R20 and R21
SendTwo:
   ; output two bytes to SPI
   out   SPDR,r20        ; send the first byte, the row - put it in SPI data register
1: in    Tmp,SPSR        ; Read SPI status register 
   sbrs  Tmp,SPIF        ; was the done bit set ?
   rjmp  1b              ; nope, keep waiting

   out   SPDR,r21        ; 2nd byte Display pattern - send it too
1: in    Tmp,SPSR        ; Read SPI status register 
   sbrs  Tmp,SPIF        ; was the done bit set ?
   rjmp  1b              ; nope, keep waiting

   ret

; ==== wait for N centi-seconds (1..255). Value in R20/tmp2, but only using the byte
delayCenti:
   push  r20             ; save registers
   push  r26             
   push  r27

   ; loop for one centisecond consuming 160000 cycles
2: ldi r26,lo8(40000)     ; 1/4 of cycles needed
   ldi r27,hi8(40000)     ; as the loop consumes 4 cycles
1:
   sbiw r26,1             ; count loops done
   brne 1b
  
   dec  r20               ; count required number
   brne 2b
  
   pop r27               ; restore and return
   pop r26
   pop r20
   ret

;====== MAIN, START HERE ====
main:

   cli                  ; we live without interrupts
  
   ; ---- Setup SPI mode
   ldi  Tmp,0B00101100  ; bit 2, 3 and 5 for Arduino pins 10, 11 and 13 as OUTPUT
   out  DDRB,Tmp        ;   the rest as INPUT
   ldi  Tmp,0B00000100  ; Bit 2/pin 10
   out  PORTB,Tmp       ;   HIGH
   ldi  Tmp,0B01010001  ; Bit 0 enable SPI, bit 5 to ?, bit 7 to ?
   out  SPCR,Tmp        ; Set these on the SPI control register

   ; ---- Setup variables (well, just some non-zero startup value for now)
   clr  Tmp
   mov  Xdir,Tmp        ; initial move left
   inc  Tmp
   mov  Ydir,Tmp        ;  ... down
   inc  Tmp
   mov  Xpos,Tmp        ; set position to 2,3
   mov  CurLin,Tmp
   inc  Tmp
   mov  Ypos,Tmp
   ldi  Tmp,20           ; Set speed to N centi seconds
   mov  Speed,Tmp        ; 
   rjmp Disp             ; start where the row is displayed
   
loop:
   ; ==== Loop ====
   ; Clear the last row used
   clr   LinBuf          ; clear buffer
   clr   LinBuf2
   rcall SendRow         ; send (CurLin should still be current)
   ; handle bounce
   ldi   Tmp,1
   cpi   Xpos,ColLft     ; reached left edge?
   brne  1f
   mov   Xdir,Tmp        ; yes, switch Xdir to 1
1: cpi   Xpos,ColRgt     ; reached right edge?
   brne  1f
   clr   Xdir            ; yes, switch Xdir to 0
1: cpi   Ypos,RowLow     ; reached top(low number) row?
   brne  1f
   mov   Ydir,Tmp        ; yes, switch Ydir to 1
1: cpi   Ypos,RowHig     ; reached bottom(high) row ?
   brne  1f
   clr   Ydir
1: ; all adjusted

   ; move to new position
   ldi   Tmp,1
   sbrs  Xdir,0          ; left or right? (bit 0 set means right, increment)
   neg   Tmp             ; want negative adjustment
   add   Xpos,Tmp        ; adjust position
   ldi   Tmp,1
   sbrs  Ydir,0          ; down or up? (bit 0 set means down, increment)
   neg   Tmp             ; want negative adjustment
   add   Ypos,Tmp        ; adjust position

Disp:
   ; display it
   mov   CurLin,Ypos     ; The row ball is on
   clr   LinBuf          ; the lit ball in the LinBuf/2
   inc   LinBuf          ;(cant ldi into LinBuf (stupid register choice))
   mov   Tmp,Xpos        ; the column ball is on
   tst   Tmp
1: breq  2f              ; have we moved it far enough to the left?
   lsl   LinBuf          ; no, give it a nudge
   rol   LinBuf2         ;   carry into next byte
   dec   Tmp             ; count
   rjmp  1b              ;       and loop
2: rcall SendRow         ; output LinBuf/2 on CurLin, ie display ball
   
   ; wait and loop
   mov   r20,Speed
   rcall delayCenti
   rjmp   loop            ; one more time !!
